"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.initialize = void 0;
var canvas_1 = require("canvas");
var index_1 = require("./index");
(0, index_1.initializeCanvas)(canvas_1.createCanvas);
function initialize() {
    (0, index_1.initializeCanvas)(canvas_1.createCanvas);
}
exports.initialize = initialize;
//# sourceMappingURL=initializeCanvas.js.map